"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.visitDocument = visitDocument;

var _HTypes = require("./HTypes");

var _HDocument = require("./HDocument");

/**
 * Traversal of a normalized document, calling
 * for each node the visitor function passed as parameter.
 *
 * The traversal is breadth first by default, unless the
 * goBreadthFirst parameter is false in which case its depthFirst
 *
 * @param {IDocumentSchema<MapsInterface, V>} docSchema
 * @param {INormalizedDocument<MapsInterface, U, V>} doc
 * @param {IVisitor<MapsInterface, U, V>} onNodeVisit
 * @param {Context} context
 */
function visitDocument(doc, onNodeVisit, {
  context,
  traversal = _HTypes.DocumentVisitTraversal.BREADTH_FIRST,
  startElement,
  typesToTraverse,
  typesToVisit
} = {}) {
  const elementType = startElement ? startElement.type : doc.rootType;
  const elementId = startElement ? startElement._id : doc.rootId;
  const traversableMap = typesToTraverse ? new Set(typesToTraverse) : undefined;
  const visitableMap = typesToVisit ? new Set(typesToVisit) : undefined;

  for (const nodesToVisit = traversal === _HTypes.DocumentVisitTraversal.BREADTH_FIRST ? breadthFirstNodes(doc, elementType, elementId, traversableMap, visitableMap) : depthFirstNodes(doc, elementType, elementId, traversableMap, visitableMap); nodesToVisit.length > 0;) {
    const [nextType, nextId] = nodesToVisit.shift();
    onNodeVisit(doc, nextType, nextId, context || {});
  }
}

function breadthFirstNodes(doc, nodeType, nodeId, typesToTraverse, typesToVisit, nodesVisited = new Set()) {
  const nodeUid = `${nodeType}:${nodeId}`;

  if (nodesVisited.has(nodeUid)) {
    return [];
  } else {
    nodesVisited.add(nodeUid);
  }

  const element = (0, _HDocument.mappedElement)(doc.maps, nodeType, nodeId);
  if (!(0, _HDocument.isParentedId)(element)) return [];
  const nodeList = typesToVisit && !typesToVisit.has(nodeType) ? [] : [[nodeType, nodeId]];
  const childrenToVisit = [];
  const nodeSchema = doc.schema.types[nodeType];

  for (const linkField in nodeSchema) {
    if (linkField === 'parentId') continue;
    const linkFieldProps = nodeSchema[linkField];

    if (Array.isArray(linkFieldProps)) {
      const {
        __schemaType
      } = linkFieldProps[0];

      if (typesToTraverse && !typesToTraverse.has(__schemaType)) {
        continue;
      }

      const fieldIds = element[linkField];

      for (const fieldId of fieldIds) {
        childrenToVisit.push([__schemaType, fieldId]);
      }
    } else {
      const {
        __schemaType
      } = linkFieldProps;

      if (typesToTraverse && !typesToTraverse.has(__schemaType)) {
        continue;
      }

      childrenToVisit.push([__schemaType, element[linkField]]);
    }
  }

  for (const [childType, childId] of childrenToVisit) {
    nodeList.push(...breadthFirstNodes(doc, childType, childId, typesToTraverse, typesToVisit, nodesVisited));
  }

  return nodeList;
}

function depthFirstNodes(doc, nodeType, nodeId, typesToTraverse, typesToVisit, nodesVisited = new Set()) {
  const nodeUid = `${nodeType}:${nodeId}`;

  if (nodesVisited.has(nodeUid)) {
    return [];
  } else {
    nodesVisited.add(nodeUid);
  }

  const nodeList = typesToVisit && !typesToVisit.has(nodeType) ? [] : [[nodeType, nodeId]];
  const element = (0, _HDocument.mappedElement)(doc.maps, nodeType, nodeId);
  if (!(0, _HDocument.isParentedId)(element)) return [];
  const nodeSchema = doc.schema.types[nodeType];

  for (const linkField in nodeSchema) {
    if (linkField === 'parentId') continue;
    const linkFieldProps = nodeSchema[linkField];

    if (Array.isArray(linkFieldProps)) {
      const {
        __schemaType
      } = linkFieldProps[0];

      if (typesToTraverse && !typesToTraverse.has(__schemaType)) {
        continue;
      }

      const fieldIds = element[linkField];

      for (const fieldId of fieldIds) {
        nodeList.unshift(...depthFirstNodes(doc, __schemaType, fieldId, typesToTraverse, typesToVisit, nodesVisited));
      }
    } else {
      const {
        __schemaType
      } = linkFieldProps;

      if (typesToTraverse && !typesToTraverse.has(__schemaType)) {
        continue;
      }

      nodeList.unshift(...depthFirstNodes(doc, __schemaType, element[linkField], typesToTraverse, typesToVisit, nodesVisited));
    }
  }

  return nodeList;
}
//# sourceMappingURL=HVisit.js.map